# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
from typing import Tuple
from datetime import datetime
from ..utils import timestamp
from ..encoding import Component, Name, ModelField, TlvModel, ContentType, BytesField,\
    SignatureInfo, TypeNumber, RepeatedField, IncludeBase, MetaInfo, VarBinaryStr,\
    get_tl_num_size, write_tl_num, parse_and_check_tl, FormalName
from ..encoding.ndn_format_0_3 import DataPacketValue


KEY_COMPONENT = Component.from_str('KEY')
SELF_COMPONENT = Component.from_str('self')


class SecurityV2TypeNumber:
    VALIDITY_PERIOD = 0xFD
    NOT_BEFORE = 0xFE
    NOT_AFTER = 0xFF
    ADDITIONAL_DESCRIPTION = 0x0102
    DESCRIPTION_ENTRY = 0x0200
    DESCRIPTION_KEY = 0x0201
    DESCRIPTION_VALUE = 0x0202

    SAFE_BAG = 0x80
    ENCRYPTED_KEY_BAG = 0x81


class DescriptionEntry(TlvModel):
    description_key = BytesField(SecurityV2TypeNumber.DESCRIPTION_KEY)
    description_value = BytesField(SecurityV2TypeNumber.DESCRIPTION_VALUE)


class AdditionalDescription(TlvModel):
    description_entry = RepeatedField(ModelField(SecurityV2TypeNumber.DESCRIPTION_ENTRY, DescriptionEntry))


class CertificateV2Extension(TlvModel):
    additional_description = ModelField(SecurityV2TypeNumber.ADDITIONAL_DESCRIPTION, AdditionalDescription)


class ValidityPeriod(TlvModel):
    not_before = BytesField(SecurityV2TypeNumber.NOT_BEFORE)
    not_after = BytesField(SecurityV2TypeNumber.NOT_AFTER)


class CertificateV2SignatureInfo(SignatureInfo, CertificateV2Extension):
    signature_info = IncludeBase(SignatureInfo)
    validity_period = ModelField(SecurityV2TypeNumber.VALIDITY_PERIOD, ValidityPeriod)
    certificate_v2_extension = IncludeBase(CertificateV2Extension)


class CertificateV2Value(DataPacketValue):
    _base = IncludeBase(DataPacketValue)
    signature_info = ModelField(TypeNumber.SIGNATURE_INFO, CertificateV2SignatureInfo, ignore_critical=True)


class SafeBag(TlvModel):
    certificate_v2 = BytesField(TypeNumber.DATA)
    # We do not use ModelField due to 2 reasons:
    # 1. The encoded length of CertificateV2 is unknown.
    # 2. Generally we already have an encoded certificate when exporting a SafeBag.
    encrypted_key_bag = BytesField(SecurityV2TypeNumber.ENCRYPTED_KEY_BAG)


def self_sign(key_name, pub_key, signer) -> Tuple[FormalName, VarBinaryStr]:
    cert_val = CertificateV2Value()
    cert_name = Name.normalize(key_name) + [SELF_COMPONENT, Component.from_version(timestamp())]
    cert_val.name = cert_name
    cert_val.content = pub_key
    cert_val.meta_info = MetaInfo(content_type=ContentType.KEY, freshness_period=3600000)
    cert_val.signature_info = CertificateV2SignatureInfo()
    cert_val.signature_info.validity_period = ValidityPeriod()
    cert_val.signature_info.validity_period.not_before = b'19700101T000000'
    end_time = datetime.utcnow()
    not_after = (f'{end_time.year+20:04}{end_time.month:02}{end_time.day:02}T'
                 f'{end_time.hour:02}{end_time.minute:02}{end_time.second:02}').encode()
    cert_val.signature_info.validity_period.not_after = not_after

    markers = {}
    cert_val._signer.set_arg(markers, signer)
    value = cert_val.encode(markers=markers)
    shrink_size = cert_val._shrink_len.get_arg(markers)
    type_len = get_tl_num_size(TypeNumber.DATA)
    size_len = get_tl_num_size(len(value) - shrink_size)
    buf = bytearray(type_len + size_len + len(value) - shrink_size)
    write_tl_num(TypeNumber.DATA, buf)
    write_tl_num(len(value) - shrink_size, buf, type_len)
    buf[type_len+size_len:] = memoryview(value)[0:len(value)-shrink_size]
    return cert_name, buf


def parse_certificate(wire) -> CertificateV2Value:
    wire = parse_and_check_tl(wire, TypeNumber.DATA)
    return CertificateV2Value.parse(wire)



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
import struct
from ..utils import timestamp, gen_nonce_64
from ..encoding import Component, Name, ModelField, TlvModel, NameField, UintField, BytesField,\
    SignatureInfo, get_tl_num_size, TypeNumber, write_tl_num, IncludeBase, parse_and_check_tl,\
    RepeatedField
from ..security import DigestSha256Signer


class Strategy(TlvModel):
    name = NameField()


class ControlParametersValue(TlvModel):
    name = NameField()
    face_id = UintField(0x69)
    uri = BytesField(0x72)
    local_uri = BytesField(0x81)
    origin = UintField(0x6f)
    cost = UintField(0x6a)
    capacity = UintField(0x83)
    count = UintField(0x84)
    base_congestion_mark_interval = UintField(0x87)
    default_congestion_threshold = UintField(0x88)
    mtu = UintField(0x89)
    flags = UintField(0x6c)
    mask = UintField(0x70)
    strategy = ModelField(0x6b, Strategy)
    expiration_period = UintField(0x6d)
    face_persistency = UintField(0x85)


class ControlParameters(TlvModel):
    cp = ModelField(0x68, ControlParametersValue)


class ControlResponse(ControlParametersValue):
    status_code = UintField(0x66)
    status_text = BytesField(0x67)
    body = IncludeBase(ControlParametersValue)


class FaceEventNotificationValue(TlvModel):
    face_event_kind = UintField(0xc1)
    face_id = UintField(0x69)
    uri = BytesField(0x72)
    local_uri = BytesField(0x81)
    face_scope = UintField(0x84)
    face_persistency = UintField(0x85)
    link_type = UintField(0x86)
    flags = UintField(0x6c)


class FaceEventNotification(TlvModel):
    event = ModelField(0xc0, FaceEventNotificationValue)


class GeneralStatus(TlvModel):
    nfd_version = BytesField(0x80)
    start_timestamp = UintField(0x81)
    current_timestamp = UintField(0x82)
    n_name_tree_entries = UintField(0x83)
    n_fib_entries = UintField(0x84)
    n_pit_entries = UintField(0x85)
    n_measurement_entries = UintField(0x86)
    n_cs_entries = UintField(0x87)
    n_in_interests = UintField(0x90)
    n_in_data = UintField(0x91)
    n_in_nacks = UintField(0x97)
    n_out_interests = UintField(0x92)
    n_out_data = UintField(0x93)
    n_out_nacks = UintField(0x98)
    n_satisfied_interests = UintField(0x99)
    n_unsatisfied_interests = UintField(0x9a)
    n_fragmentation_errors = UintField(0xc8)
    n_out_over_mtu = UintField(0xc9)
    n_in_lp_invalid = UintField(0xca)
    n_reassembly_timeouts = UintField(0xcb)
    n_in_net_invalid = UintField(0xcc)
    n_acknowledged = UintField(0xcd)
    n_retransmitted = UintField(0xce)
    n_retx_exhausted = UintField(0xcf)
    n_congestion_marked = UintField(0xd0)


class FaceStatus(TlvModel):
    face_id = UintField(0x69)
    uri = BytesField(0x72)
    local_uri = BytesField(0x81)
    expiration_period = UintField(0x6d)
    face_scope = UintField(0x84)
    face_persistency = UintField(0x85)
    link_type = UintField(0x86)
    base_congestion_mark_interval = UintField(0x87)
    default_congestion_threshold = UintField(0x88)
    mtu = UintField(0x89)
    n_in_interests = UintField(0x90)
    n_in_data = UintField(0x91)
    n_in_nacks = UintField(0x97)
    n_out_interests = UintField(0x92)
    n_out_data = UintField(0x93)
    n_out_nacks = UintField(0x98)
    n_in_bytes = UintField(0x94)
    n_out_bytes = UintField(0x95)
    flags = UintField(0x6c)


class FaceStatusMsg(TlvModel):
    face_status = RepeatedField(ModelField(0x80, FaceStatus))


class FaceQueryFilterValue(TlvModel):
    face_id = UintField(0x69)
    uri_scheme = BytesField(0x83)
    uri = BytesField(0x72)
    local_uri = BytesField(0x81)
    face_scope = UintField(0x84)
    face_persistency = UintField(0x85)
    link_type = UintField(0x86)


class FaceQueryFilter(TlvModel):
    face_query_filter = ModelField(0x96, FaceQueryFilterValue)


class Route(TlvModel):
    face_id = UintField(0x69)
    origin = UintField(0x6f)
    cost = UintField(0x6a)
    flags = UintField(0x6c)
    expiration_period = UintField(0x6d)


class RibEntry(TlvModel):
    name = NameField()
    routes = RepeatedField(ModelField(0x81, Route))


class RibStatus(TlvModel):
    entries = RepeatedField(ModelField(0x80, RibEntry))


class NextHopRecord(TlvModel):
    face_id = UintField(0x69)
    cost = UintField(0x6a)


class FibEntry(TlvModel):
    name = NameField()
    next_hop_records = RepeatedField(ModelField(0x81, NextHopRecord))


class FibStatus(TlvModel):
    entries = RepeatedField(ModelField(0x80, FibEntry))


class StrategyChoice(TlvModel):
    name = NameField()
    strategy = ModelField(0x6b, Strategy)


class StrategyChoiceMsg(TlvModel):
    strategy_choices = RepeatedField(ModelField(0x80, StrategyChoice))


class CsInfo(TlvModel):
    capacity = UintField(0x83)
    flags = UintField(0x6c)
    n_cs_entries = UintField(0x87)
    n_hits = UintField(0x81)
    n_misses = UintField(0x82)


def make_command(module, command, **kwargs):
    ret = Name.from_str(f"/localhost/nfd/{module}/{command}")

    # Command parameters
    cp = ControlParameters()
    cp.cp = ControlParametersValue()
    for k, v in kwargs.items():
        if k == 'strategy':
            cp.cp.strategy = Strategy()
            cp.cp.strategy.name = v
        else:
            setattr(cp.cp, k, v)
    ret.append(Component.from_bytes(cp.encode()))

    # Timestamp and nonce
    ret.append(Component.from_bytes(struct.pack('!Q', timestamp())))
    ret.append(Component.from_bytes(struct.pack('!Q', gen_nonce_64())))

    # SignatureInfo
    signer = DigestSha256Signer()
    sig_info = SignatureInfo()
    signer.write_signature_info(sig_info)
    buf = sig_info.encode()
    ret.append(Component.from_bytes(bytes([TypeNumber.SIGNATURE_INFO, len(buf)]) + buf))

    # SignatureValue
    sig_size = signer.get_signature_value_size()
    tlv_length = 1 + get_tl_num_size(sig_size) + sig_size
    buf = bytearray(tlv_length)
    buf[0] = TypeNumber.SIGNATURE_VALUE
    offset = 1 + write_tl_num(sig_size, buf, 1)
    signer.write_signature_value(memoryview(buf)[offset:], ret)
    ret.append(Component.from_bytes(buf))

    return ret


def parse_response(buf):
    buf = parse_and_check_tl(memoryview(buf), 0x65)
    cr = ControlResponse.parse(buf)
    ret = {}
    for k in ControlResponse._encoded_fields:
        val = getattr(cr, k.name)
        if isinstance(val, memoryview):
            val = bytes(val)
        ret[k.name] = val
    return ret



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
import os
from base64 import b64decode, b64encode
from hashlib import sha256
from typing import Tuple
from Cryptodome.PublicKey import RSA, ECC
from ...encoding import Signer, NonStrictName, Name, BinaryStr, FormalName
from ..signer.sha256_rsa_signer import Sha256WithRsaSigner
from ..signer.sha256_ecdsa_signer import Sha256WithEcdsaSigner
from .tpm import Tpm


class TpmFile(Tpm):
    path: str

    def __init__(self, path):
        self.path = path

    @staticmethod
    def _to_file_name(key_name: bytes):
        return sha256(key_name).digest().hex() + '.privkey'

    @staticmethod
    def _base64_newline(src: bytes):
        return b'\n'.join(src[i*64:i*64+64] for i in range((len(src) + 63) // 64))

    def get_signer(self, key_name: NonStrictName) -> Signer:
        key_name = Name.to_bytes(key_name)
        file_name = os.path.join(self.path, self._to_file_name(key_name))
        if not os.path.exists(file_name):
            raise KeyError(key_name)
        with open(file_name, 'rb') as f:
            key_b64 = f.read()
        key_der = b64decode(key_b64)
        for signer in [Sha256WithRsaSigner, Sha256WithEcdsaSigner]:
            try:
                return signer(key_name, key_der)
            except ValueError:
                pass
        raise ValueError('Key format is not supported')

    def key_exist(self, key_name: FormalName) -> bool:
        key_name = Name.encode(key_name)
        file_name = os.path.join(self.path, self._to_file_name(key_name))
        return os.path.exists(file_name)

    def save_key(self, key_name: FormalName, key_der: BinaryStr):
        key_name = Name.encode(key_name)
        key_b64 = self._base64_newline(b64encode(key_der))
        file_name = os.path.join(self.path, self._to_file_name(key_name))
        with open(file_name, 'wb') as f:
            f.write(key_b64)

    def delete_key(self, key_name: FormalName):
        key_name = Name.encode(key_name)
        file_name = os.path.join(self.path, self._to_file_name(key_name))
        try:
            os.remove(file_name)
        except FileNotFoundError:
            pass

    def generate_key(self, id_name: FormalName, key_type: str = 'rsa', **kwargs) -> Tuple[FormalName, BinaryStr]:
        if key_type == 'rsa':
            siz = kwargs.pop('key_size', 2048)
            pri_key = RSA.generate(siz)
            pub_key = pri_key.publickey().export_key(format='DER')
            key_der = pri_key.export_key(format='DER', pkcs=1)
        elif key_type == 'ec':
            siz = kwargs.pop('key_size', 256)
            pri_key = ECC.generate(curve=f'P-{siz}')
            pub_key = bytes(pri_key.public_key().export_key(format='DER'))
            key_der = pri_key.export_key(format='DER', use_pkcs8=False)
        else:
            raise ValueError(f'Unsupported key type {key_type}')
        key_name = self.construct_key_name(id_name, pub_key, **kwargs)
        self.save_key(key_name, key_der)
        return key_name, pub_key



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2021 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
from typing import Optional
from ..encoding import NonStrictName, Name, BinaryStr, InterestParam, FormalName
from ..types import Validator, Route
from ..name_tree import NameTrie, PrefixTreeNode


class Dispatcher:
    """
    An Interest dispatcher that helps a producer application further dispatches Interests under some route.
    """

    _tree: NameTrie = None

    def __init__(self):
        self._tree = NameTrie()

    def register(self, name: NonStrictName, func: Route):
        """
        Register a callback function. This will not register an NDN route.

        :param name: the name prefix.
        :param func: the callback function.
        :raises ValueError: the name prefix is already registered.
        """
        name = Name.normalize(name)
        node = self._tree.setdefault(name, PrefixTreeNode())
        if node.callback:
            raise ValueError(f'Duplicated registration: {Name.to_str(name)}')
        node.callback = func

    def unregister(self, name: NonStrictName):
        """
        Unregister a callback function.

        :param name: the name prefix.
        """
        name = Name.normalize(name)
        del self._tree[name]

    def dispatch(self, name: FormalName, param: InterestParam, app_param: Optional[BinaryStr]) -> bool:
        """
        Dispatch the Interest to registered callbacks using longest match.

        :return: ``True`` if the Interest is dispatched to some callbacks.
        """
        trie_step = self._tree.longest_prefix(name)
        if not trie_step:
            return False
        trie_step.value.callback(name, param, app_param)
        return True



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
from ..encoding import NonStrictName, Name, Component
from ..app import NDNApp
from ..types import InterestTimeout


async def segment_fetcher(app: NDNApp, name: NonStrictName, timeout=4000, retry_times=3,
                          validator=None, must_be_fresh=True):
    """
    An async-generator to fetch a segmented object. Interests are issued one by one.

    :param app: NDN Application
    :param name: Name prefix of Data
    :param timeout: Timeout value, in milliseconds
    :param retry_times: Times for retry
    :param validator: Validator
    :param must_be_fresh: MustBeFresh field of Interest
    :return: Data segments in order.
    """
    async def retry(first):
        nonlocal name
        trial_times = 0
        while True:
            future = app.express_interest(name, validator=validator, can_be_prefix=first,
                                          must_be_fresh=must_be_fresh, lifetime=timeout)
            try:
                return await future
            except InterestTimeout:
                trial_times += 1
                if trial_times >= retry_times:
                    raise

    name = Name.normalize(name)
    # First Interest
    name, meta, content = await retry(True)
    # If it's not segmented
    if Component.get_type(name[-1]) != Component.TYPE_SEGMENT:
        yield content
        return
    # If it's segmented
    if Component.to_number(name[-1]) == 0:
        yield content
        if meta.final_block_id == name[-1]:
            return
        seg_no = 1
    else:
        # If it's not segment 0, starting from 0
        seg_no = 0
    # Following Interests
    while True:
        name[-1] = Component.from_segment(seg_no)
        name, meta, content = await retry(False)
        yield content
        if meta.final_block_id == name[-1]:
            return
        seg_no += 1



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
import asyncio as aio
from typing import Dict, Any, Type, Optional
from dataclasses import dataclass
from ..encoding import is_binary_str, FormalName, NonStrictName, Name, Component, \
    SignaturePtrs, InterestParam, BinaryStr, MetaInfo, parse_data, TypeNumber
from ..app import NDNApp
from ..security import sha256_digest_checker, DigestSha256Signer
from ..utils import gen_nonce
from .util import norm_pattern
from . import policy


class NodeExistsError(Exception):
    """
    Raised when trying to create a node which already exists.
    """
    pattern: str

    def __init__(self, pattern: str):
        self.pattern = pattern


class LocalResourceNotExistError(Exception):
    """
    Raised when trying to fetch a local resource that does not exist.
    Used only when :class:`LocalOnly` is attached to the node.
    """
    name: FormalName

    def __init__(self, name: FormalName):
        self.name = name


class Node:
    """
    Node represents a node in the static namespace tree.

    :ivar policies: policies attached to this node
    :vartype policies: Dict[Type[policy.Policy], policy.Policy]
    :ivar prefix: the prefix of the root node of the tree. Generally not set for other nodes.
    :vartype prefix: :any:`FormalName`
    :ivar ~.app: the :any:`NDNApp` this static tree is attached to. Only available at the root.
    :vartype ~.app: Optional[NDNApp]
    """
    policies: Dict[Type[policy.Policy], policy.Policy]
    prefix: FormalName
    app: Optional[NDNApp]

    def __init__(self, parent=None):
        self.parent = parent
        # Efficiency is not considered at this draft
        self.children = {}
        self.matches = {}
        self.policies = {}
        self.prefix = []
        self.app = None

    # def make_namespace(self, prefix: NonStrictName):
    #     ret = Node()
    #     self[prefix] = ret
    #     return ret

    # ====== Functions operating on children ======

    def exist(self, key):
        """
        If it has a child with specified name component or nme pattern.

        :param key: a name component (bytes) or a patten (tuple).
        :return: whether the child node exists
        """
        if is_binary_str(key):
            return bytes(key) in self.children
        else:
            return key[:2] in self.matches

    def _get(self, key):
        if is_binary_str(key):
            return self.children[bytes(key)]
        else:
            return self.matches[key[:2]][1]

    def _set(self, key, val):
        if is_binary_str(key):
            self.children[bytes(key)] = val
        else:
            self.matches[key[:2]] = (key[2], val)
        return val

    def __getitem__(self, key: str):
        """
        Get a node in the subtree rooted at this node.
        If any node on the path does not exist, this function will create it.

        :param key: the path from this node to the destination. Can contain both components and patterns.
        :return: the desinated node.
        """
        key_lst = norm_pattern(key)
        cur = self
        for k in key_lst:
            try:
                cur = cur._get(k)
            except KeyError:
                cur = cur._set(k, Node(cur))
        return cur

    def __setitem__(self, key: str, value):
        """
        Set a node in the subtree rooted at this node.
        If any node on the path does not exist, this function will create it.

        :param key: the path from this node to the destination. Can contain both components and patterns.
        :param value: the destinated node.
        :return: the same as ``value``
        """
        key_lst = norm_pattern(key)
        cur = self
        for k in key_lst[:-1]:
            try:
                cur = cur._get(k)
            except KeyError:
                cur = cur._set(k, Node(cur))
        if cur.exist(key_lst[-1]):
            raise NodeExistsError(key)
        cur._set(key_lst[-1], value)
        value.parent = cur
        return value

    def _match_step(self, comp: bytes, env, policies):
        policies.update(self.policies)
        chd = self.children.get(comp, None)
        if chd is not None:
            return chd
        typ = Component.get_type(comp)
        match = self.matches.get((0, typ), None)
        if match is not None:
            env[match[0]] = Component.get_value(comp)
            return match[1]
        else:
            return None

    def match(self, name: NonStrictName):
        """
        Start from this node, go the path that matches with the name,
        and return the node it reaches when it cannot go further.

        :param name: an NDN name.
        :return: a :class:`MatchedNode`, which contains the destination node and variables matched.
        """
        if self.parent is not None:
            raise ValueError('Node.match() should be called from root')
        env = {}
        policies = {}
        cur = self
        name = Name.normalize(name)
        if self.prefix:
            if len(name) < len(self.prefix) or name[:len(self.prefix)] != self.prefix:
                raise ValueError(f'The name f{Name.to_str(name)} does not match with '
                                 f'the prefix of this node {Name.to_str(self.prefix)}')
            pos = len(self.prefix)
        else:
            pos = 0
        while pos < len(name):
            nxt = cur._match_step(bytes(name[pos]), env, policies)
            if not nxt:
                break
            else:
                cur = nxt
                pos += 1
        if pos is None:
            pos = len(name)
        policies.update(cur.policies)
        return MatchedNode(root=self, node=cur, name=name, pos=pos, env=env, policies=policies)

    # TODO: Apply

    # ====== Functions operating on policies ======

    def get_policy(self, typ: Type[policy.Policy]):
        """
        Get the policy of specified type that applies to this node.
        It can be attached to this node or a parent of this node.

        :param typ: a policy type
        :return: the policy. None if there does not exist one.
        """
        ret = None
        cur = self
        while ret is None and cur is not None:
            ret = cur.policies.get(typ, None)
            cur = cur.parent
        return ret

    def set_policy(self, typ: Type[policy.Policy], value: policy.Policy):
        """
        Attach a policy to this node.

        :param typ: the policy type.
        :param value: the policy to be attached to this node.
        """
        if not isinstance(value, typ):
            raise TypeError(f'The policy {value} is not of type {typ}')
        self.policies[typ] = value
        value.node = self

    # ====== Functions on registration  ======

    async def attach(self, app: NDNApp, prefix: NonStrictName):
        r"""
        Attach this node to a specified :any:`NDNApp`, register all name prefixes.
        This node becomes the root node of the application static tree.
        ``prefix`` is the prefix of the tree, which will be prepended to all names under this tree.
        For example, if ``prefix='/a/blog'``, then the node with path ``/articles`` from this node
        will become ``/a/blog/articles``.

        .. warning::

            The way to register prefixes is still under discussion.
            Currently, we register the nodes that we can reach without going through a pattern.
            Also, there is no ``detach`` function yet, and no means to change the static tree after it's attached.

        :param app: the :any:`NDNApp` to be attached to.
        :param prefix: the prefix of the static tree.
        :return: whether succeeded or not.
        """
        prefix = Name.normalize(prefix)
        self.app = app
        return await self.on_register(self, app, prefix, cached=False)

    # async def detach(self, app: NDNApp):
    #     raise NotImplementedError('TODO: Not supported yet. Please reset NDNApp.')

    async def on_register(self, root, app: NDNApp, prefix: FormalName, cached: bool):
        """
        Called when the root node ``root`` is attached to ``app``, and the :meth:`attach` wants to
        register prefixed under the subtree rooted at this node.

        :param root: the root of the static tree.
        :param app: the :any:`NDNApp` to be attached to.
        :param prefix: the prefix of the static tree.
        :param cached: If there is a cache policy that applies to this node.
        :return: whether succeeded or not.
        """
        # If there is a register policy
        if policy.Register in self.policies:
            return await app.register(prefix, root._on_interest_root, root._int_validator, True)
        # If it is cached with a match or being leaf
        cached = cached or policy.Cache in self.policies
        if cached:
            if self.matches or not self.children:
                return await app.register(prefix, root._on_interest_root, root._int_validator, True)
        # O/w enumerate its children
        for comp, chd in self.children.items():
            if not await chd.on_register(root, app, prefix + [comp], cached=cached):
                return False
        return True

    async def _int_validator(self, name: FormalName, sig_ptrs: SignaturePtrs) -> bool:
        match = self.match(name)
        validate_policy = match.policies.get(policy.InterestValidator, None)
        if validate_policy is None:
            return await sha256_digest_checker(name, sig_ptrs)
        if isinstance(validate_policy, policy.InterestValidator):
            return await validate_policy.validate(match, sig_ptrs)
        else:
            raise TypeError(f'The InterestValidator policy is of wrong type. Name={Name.to_str(name)}')

    def _on_interest_root(self, name: FormalName, param: InterestParam,
                          app_param: Optional[BinaryStr], raw_packet: BinaryStr):
        match = self.match(name)
        aio.ensure_future(match.on_interest(param, app_param, raw_packet))

    # ====== Functions on Interest & Data processing (For overriding)  ======

    async def process_int(self, match, param: InterestParam, app_param: Optional[BinaryStr], raw_packet: BinaryStr):
        """
        Processing an incoming Interest packet. Specific node type can override this function to have customized
        processing pipeline.

        .. note::

            This function will not be called if the Interest packet is satisfied with a cached Data packet.

        :param match: the matched node object of this node.
        :param param: the parameters of the Interest packet.
        :param app_param: the ApplicationParameters of the Interest packet.
        :param raw_packet: the raw Interest packet.
        """
        pass

    async def process_data(self, match, meta_info: MetaInfo, content: Optional[BinaryStr], raw_packet: BinaryStr):
        """
        Processing an incoming Data packet. Specific node type can override this function to have customized
        processing pipeline. By default it returns the content.

        :param match: the matched node object of this node.
        :param meta_info: the MetaInfo of the Data packet.
        :param content: the content of the Data packet.
        :param raw_packet: the raw Data packet.
        :return: a tuple, whose first element is data content after processing, and second is a ``dict[str, Any]``
            containing metadata.
        """
        # Override this function to customize the processing
        meta_data = {
            **match.env,
            'content_type': meta_info.content_type,
            'freshness_period': meta_info.freshness_period,
            'final_block_id': meta_info.final_block_id
        }
        return content, meta_data

    async def need(self, match, **kwargs):
        """
        Consume an object corresponding to this node. Specific node type can override this function to have customized
        processing pipeline. For example, a SegmentedNode can do reassembly here.
        By default it sends an Interest packet to fetch a Data.

        :param match: the matched node object of this node.
        :param kwargs: other arguments from user input.
        :return: This is defined by the node type. By default it returns what :meth:`process_data` returns.
            That is, a tuple of contect and metadata dict.
        """
        return await match.express(**kwargs)

    async def provide(self, match, content, **kwargs):
        """
        Produce an object corresponding to this node, and make all generated Data packets available.
        Specific node type can override this function to have customized processing pipeline.
        For example, a SegmentedNode can do segmentation here.
        By default it makes a Data packet out of content and put it into the cache.

        :param match: the matched node object of this node.
        :param content: the content of the object.
        :param kwargs: other arguments from user input.
        """
        return await match.put_data(content, **kwargs)


@dataclass
class MatchedNode:
    r"""
    MatchedNode represents a matched static tree node.
    That is, a node with all name patterns on the path from the root to it assigned to some value.
    For example, if the tree contains a node N on the path ``/a/<b>/<c>``,
    and the user use the Name ``/a/x/y`` to match,
    then a matched node (N, {'b': 'x', 'c': 'y'}) will be returned.

    :ivar root: the root of the static tree.
    :vartype root: Node
    :ivar node: the matched node of the static tree.
    :vartype node: Node
    :ivar name: the name used to match.
    :vartype name: :any:`FormalName`
    :ivar pos: an integer indicating the length the name is matched. Generally, it equals the length of ``name``.
    :vartype pos: int
    :ivar env: a dict containing the value all pattern variables matched on the path.
    :vartype env: Dict[str, Any]
    :ivar policies: a dict collecting all policies that apply to this node.
        For each type of policy, the one attached on the nearst ancestor is collected here.
    :vartype policies: Dict[Type[policy.Policy], policy.Policy]
    """
    root: Node
    node: Node
    name: FormalName
    pos: int
    env: Dict[str, Any]
    policies: Dict[Type[policy.Policy], policy.Policy]

    def finer_match(self, new_name: FormalName):
        """
        Do a finer match based on current match. ``new_name`` must include current ``name`` as its prefix.
        For example, if the current match name is ``/a/b`` and we want to get the matched node for ``/a/b/c``,
        then we can call finer_match with ``/a/b/c``.

        :param new_name: the new name to be matched. Must include current ``name`` as its prefix.
        :return: the new matched node.
        """
        name_len = len(self.name)
        if self.pos < name_len:
            # match = self.finer_match(data_name[name_len:])
            return MatchedNode(root=self.root, node=self.node, name=new_name, pos=self.pos,
                               env=self.env, policies=self.policies)

        env = self.env.copy()
        policies = self.policies.copy()
        pos = None
        cur = self.node
        for i in range(name_len, len(new_name)):
            comp = new_name[i]
            nxt = cur._match_step(bytes(comp), env, policies)
            if not nxt:
                pos = i
                break
            else:
                cur = nxt
        if pos is None:
            pos = len(new_name)
        policies.update(cur.policies)
        return MatchedNode(root=self.root, node=cur, name=new_name, pos=pos, env=env, policies=policies)

    async def on_interest(self, param: InterestParam, app_param: Optional[BinaryStr], raw_packet: BinaryStr):
        """
        Called when an Interest packet comes.
        It looks up the cache and returns a Data packet if it exists.
        Otherwise, it decrypts ApplicationParameters and calls the node's ``process_int`` function.

        :param param: the parameters of the incoming Interest.
        :param app_param: the ApplicationParameters of the Interest.
        :param raw_packet: the raw Interest packet.
        """
        # Cache search
        cache_policy = self.policies.get(policy.Cache, None)
        if cache_policy and isinstance(cache_policy, policy.Cache):
            data_raw = await cache_policy.search(self, self.name, param)
            if data_raw is not None:
                self.root.app.put_raw_packet(data_raw)
                return
        # By design, we do not cache Interest
        # Decrypt app_param
        if app_param:
            ac_policy = self.policies.get(policy.InterestEncryption, None)
            if ac_policy and isinstance(ac_policy, policy.InterestEncryption):
                app_param = await ac_policy.decrypt(self, app_param)
        # Process Interest
        await self.node.process_int(self, param, app_param, raw_packet)

    async def on_data(self, meta_info: MetaInfo, content: Optional[BinaryStr], raw_packet: BinaryStr):
        """
        Called when a Data packet comes.
        It saves the Data packet into the cache, decrypts the content, and calls
        the node's ``process_data`` function.

        :param meta_info: the MetaInfo of the incoming Data packet.
        :param content: the content of the Data.
        :param raw_packet: the raw Data packet.
        :return: whatever ``process_data`` returns.
        """
        # Cache save
        if policy.LocalOnly not in self.policies:
            cache_policy = self.policies.get(policy.Cache, None)
            if cache_policy and isinstance(cache_policy, policy.Cache):
                # aio.ensure_future(cache_policy.save(self, self.name, raw_packet))
                # self.name may change after this time point, so we have to wait until its finish
                await cache_policy.save(self, self.name, raw_packet)
        # Decrypt content
        if content is not None:
            ac_policy = self.policies.get(policy.DataEncryption, None)
            if ac_policy and isinstance(ac_policy, policy.DataEncryption):
                content = await ac_policy.decrypt(self, content)
        # Process Data
        return await self.node.process_data(self, meta_info, content, raw_packet)

    async def express(self, app_param: Optional[BinaryStr] = None, **kwargs):
        """
        Try to fetch the data, called by the node's need function.
        It will search the local cache, and examines the local resource.
        If the corresponding Data cannot be found in the two places,
        it encrypts the app_param and expresses the Interest.

        .. note::

            This function only sends out an Interest packet when the Data is not cached locally.

        :param app_param: the ApplicationParameter of the Interest.
        :param kwargs: other parameters of the Interest.
        :return: whatever ``process_data`` returns.
            Generally this function is only called at the default node,
            so the return value is a tuple of the content and a dict containing metadata.
        """
        if 'nonce' not in kwargs:
            kwargs['nonce'] = gen_nonce()
        param = InterestParam.from_dict(kwargs)

        # Cache search
        cache_policy = self.policies.get(policy.Cache, None)
        if cache_policy and isinstance(cache_policy, policy.Cache):
            data_raw = await cache_policy.search(self, self.name, param)
            if data_raw is not None:
                with_tl = (data_raw[0] == TypeNumber.DATA)
                data_name, meta_info, content, _ = parse_data(data_raw, with_tl=with_tl)
                return await self.finer_match(data_name).on_data(meta_info, content, data_raw)
        # Local only?
        local_policy = self.policies.get(policy.LocalOnly, None)
        if local_policy:
            raise LocalResourceNotExistError(self.name)
        # Encrypt app_param
        if app_param is not None:
            ac_policy = self.policies.get(policy.InterestEncryption, None)
            if ac_policy and isinstance(ac_policy, policy.InterestEncryption):
                app_param = await ac_policy.encrypt(self, app_param)
        # Get validator TODO: How can we pass information out?
        validate_policy = self.policies.get(policy.DataValidator, None)
        if validate_policy and isinstance(validate_policy, policy.DataValidator):
            validator = validate_policy.get_validator(self)
        else:
            validator = sha256_digest_checker  # Change this if possible
        # Get signer
        signer_policy = self.policies.get(policy.InterestSigning, None)
        if signer_policy and isinstance(signer_policy, policy.InterestSigning):
            signer = signer_policy.get_signer(self)
        elif app_param is not None:
            signer = DigestSha256Signer()
        else:
            signer = None
        # Express interest
        data = await self.root.app.express_interest(self.name, app_param, validator, need_raw_packet=True,
                                                    interest_param=param, signer=signer)
        data_name, meta_info, content, data_raw = data
        return await self.finer_match(data_name).on_data(meta_info, content, data_raw)

    def need(self, **kwargs):
        """
        Consume an object corresponding to this node. Specific node type may have customized
        processing pipeline. For example, a SegmentedNode can do reassembly here.
        By default it sends an Interest packet to fetch a Data.

        MatchedNode's ``need`` simply calls the node's ``need`` function.

        :param kwargs: arguments from user input.
        :return: the object needed, whose format is defined by specific node type.
            By default, it returns a tuple of the content and a dict of metadata.
        """
        return self.node.need(self, **kwargs)

    def provide(self, content, **kwargs):
        """
        Produce an object corresponding to this node, and make all generated Data packets available.
        Specific node type may have customized processing pipeline.
        For example, a SegmentedNode can do segmentation here.
        By default it makes a Data packet out of content and put it into the cache.

        MatchedNode's ``provide`` simply calls the node's ``provide`` function.

        :param content: the content of the object.
        :param kwargs: other arguments from user input. Defined by specific node type.
        """
        return self.node.provide(self, content, **kwargs)

    async def put_data(self, content: Optional[BinaryStr] = None, send_packet: bool = False, **kwargs):
        """
        Generate the Data packet out of content.
        This function encrypts the content, encodes and signs the packet, saves it into the cache,
        and optionally sends it to the face.
        This function is called by the node's ``provide`` function.

        :param content: the Data content.
        :param send_packet: whether sends the Data packet to the face.
        :param kwargs: other arguments generating the Data packet.
        """
        meta_info = MetaInfo.from_dict(kwargs)
        data_name = self.name
        # Encrypt content
        if content is not None:
            ac_policy = self.policies.get(policy.DataEncryption, None)
            if ac_policy and isinstance(ac_policy, policy.DataEncryption):
                content = await ac_policy.encrypt(self, content)
        # Get signer
        signer_policy = self.policies.get(policy.DataSigning, None)
        if signer_policy and isinstance(signer_policy, policy.DataSigning):
            signer = signer_policy.get_signer(self)
        else:
            signer = self.root.app.keychain.get_signer(kwargs)
        # Prepare Data packet
        raw_packet = self.root.app.prepare_data(data_name, content, meta_info=meta_info, signer=signer)
        # Cache save
        cache_policy = self.policies.get(policy.Cache, None)
        if cache_policy and isinstance(cache_policy, policy.Cache):
            # aio.ensure_future(cache_policy.save(self, self.name, raw_packet))
            await cache_policy.save(self, self.name, raw_packet)
        # face.put
        if send_packet:
            self.root.app.put_raw_packet(raw_packet)

    def app(self) -> NDNApp:
        """
        The :any:`NDNApp` the static tree is attached to.

        :return: the :any:`NDNApp`.
        """
        return self.root.app



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
"""
Name module is a collection of functions processing NDN Names.
"""
from functools import reduce
from typing import List, Optional, Iterable
from . import Component
from ..tlv_type import BinaryStr, VarBinaryStr, FormalName, NonStrictName, is_binary_str
from ..tlv_var import write_tl_num, parse_tl_num, get_tl_num_size


TYPE_NAME = 0x07
"""The TLV type of NDN Name."""


def from_str(val: str) -> List[bytearray]:
    r"""
    Construct a Name from a URI string.

    :param val: URI string. Character out of :const:`ndn.encoding.name.Component.CHARSET` will be
        escaped automatically. Leading and tailing ``'/'`` will be removed.

        .. note::
            Additional periods are not allowed here. To create a zero-size Component, use two slashes
            to surround it. Also, there should be no scheme identifier and authority component in the
            URI.

    :return: :any:`FormalName`.

    :examples:
        >>> from ndn.encoding.name import Name
        >>> Name.from_str("example/name")
        [bytearray(b'\x08\x07example'), bytearray(b'\x08\x04name')]

        >>> Name.from_str("/a//32=b/")
        [bytearray(b'\x08\x01a'), bytearray(b'\x08\x00'), bytearray(b'\x20\x01b')]

        >>> Name.from_str('/a/../b')
        [bytearray(b'\x08\x01a'), bytearray(b'\x08\x02..'), bytearray(b'\x08\x01b')]
    """
    cnt_slash = 0
    if val.startswith('/'):
        val = val[1:]
        cnt_slash += 1
    if val.endswith('/'):
        val = val[:-1]
        cnt_slash += 1
    if not val and cnt_slash <= 1:
        return []
    compstrs = val.split('/')
    return [Component.from_str(Component.escape_str(comp)) for comp in compstrs]


def to_str(name: NonStrictName) -> str:
    r"""
    Convert an NDN Name to a URI string.

    :param name: the input NDN Name.
    :type name: :any:`NonStrictName`
    :return: the URI.

    :examples:
        >>> from ndn.encoding.name import Name
        >>> Name.to_str('Σπυρίδων')
        '/%CE%A3%CF%80%CF%85%CF%81%CE%AF%CE%B4%CF%89%CE%BD'
    """
    name = normalize(name)
    ret = '/' + '/'.join(Component.to_str(comp) for comp in name)
    if name and name[-1] == b'\x08\x00':
        ret += '/'
    return ret


def from_bytes(buf: BinaryStr) -> FormalName:
    r"""
    Decode the Name from its TLV encoded form.

    :param buf: encoded Name.
    :return: Decoded Name.
    :rtype: :any:`FormalName`

    :raises ValueError: if the Type is not :any:`TYPE_NAME` .
    """
    return decode(buf)[0]


def to_bytes(name: NonStrictName) -> bytes:
    r"""
    Encode a Name via TLV encoding.

    :param name: Name to encode.
    :type name: :any:`NonStrictName`
    :return: Encoded Name.
    """
    if not is_binary_str(name):
        name = encode(normalize(name))
    return bytes(name)


def is_prefix(lhs: NonStrictName, rhs: NonStrictName) -> bool:
    r"""
    Test if a Name is a prefix of another Name.

    :param lhs: prefix to be tested.
    :type lhs: :any:`NonStrictName`
    :param rhs: full name to test on.
    :type rhs: :any:`NonStrictName`
    :return: ``True`` if ``lhs`` is a prefix of ``rhs``.
    """
    lhs = normalize(lhs)
    rhs = normalize(rhs)
    left_len = len(lhs)
    return left_len <= len(rhs) and lhs == rhs[:left_len]


def encoded_length(name: FormalName) -> int:
    length = reduce(lambda x, y: x + len(y), name, 0)
    size_typ = 1
    size_len = get_tl_num_size(length)
    return length + size_typ + size_len


def encode(name: FormalName, buf: Optional[VarBinaryStr] = None, offset: int = 0) -> VarBinaryStr:
    length = reduce(lambda x, y: x + len(y), name, 0)
    size_typ = 1
    size_len = get_tl_num_size(length)

    if not buf:
        buf = bytearray(length + size_typ + size_len)
    else:
        if len(buf) < length + size_typ + size_len + offset:
            raise IndexError('buffer overflow')

    offset += write_tl_num(TYPE_NAME, buf, offset)
    offset += write_tl_num(length, buf, offset)
    for comp in name:
        buf[offset:offset+len(comp)] = comp
        offset += len(comp)
    return buf


def decode(buf: BinaryStr, offset: int = 0) -> (List[memoryview], int):
    buf = memoryview(buf)
    origin_offset = offset

    typ, size_typ = parse_tl_num(buf, offset)
    offset += size_typ
    if typ != TYPE_NAME:
        raise ValueError(f'the Type of {buf} is not Name')

    length, size_len = parse_tl_num(buf, offset)
    offset += size_len
    if length > len(buf) - offset:
        raise IndexError('buffer overflow')

    ret = []
    while length > 0:
        st = offset
        _, size_typ_comp = parse_tl_num(buf, offset)
        offset += size_typ_comp
        len_comp, size_len_comp = parse_tl_num(buf, offset)
        offset += size_len_comp + len_comp
        ret.append(buf[st:offset])
        length -= (offset - st)

    return ret, offset - origin_offset


def normalize(name: NonStrictName) -> FormalName:
    r"""
    Convert a NonStrictName to a FormalName.
    If name is a binary string, decode it.
    If name is a str, encode it into FormalName.
    If name is a list, encode all str elements into Components.

    :param name: the NonStrictName.
    :type name: :any:`NonStrictName`
    :return: the FormalName. It may be a swallow copy of name.
    :rtype: :any:`FormalName`

    :raises TypeError: if the name or one of its element has a unrecognized type.

    :examples:
        >>> from ndn.encoding.name import Name
        >>> Name.normalize(f'{i}' for i in range(3))
        [bytearray(b'\x08\x010'), bytearray(b'\x08\x011'), bytearray(b'\x08\x012')]

        >>> Name.normalize(['Алек', b'%\x01\x00'])
        [bytearray(b'\x08\x08\xd0\x90\xd0\xbb\xd0\xb5\xd0\xba'), b'%\x01\x00']
    """
    if is_binary_str(name):
        return decode(name)[0]
    elif isinstance(name, str):
        return from_str(name)
    elif not isinstance(name, Iterable):
        raise TypeError('invalid type for name')
    ret = list(name)
    for i, comp in enumerate(ret):
        if isinstance(comp, str):
            ret[i] = Component.from_str(Component.escape_str(comp))
        elif not is_binary_str(comp):
            raise TypeError('invalid type for name component')
    return ret



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
import sys
from ctypes import cdll, c_void_p, c_ubyte, POINTER, c_int32, c_ulong, c_uint16
if sys.platform == 'darwin':
    from ..contrib.cocoapy import cf, CFIndex, CFRange, CFAllocatorRef


class OsxSec(object):
    __instance = None

    def __new__(cls):
        if OsxSec.__instance is None:
            OsxSec.__instance = object.__new__(cls)
        return OsxSec.__instance

    def __init__(self):
        if len(self.__dict__) > 0:
            return
        self.security = cdll.LoadLibrary(
            "/System/Library/Frameworks/Security.framework/Versions/Current/Security")

        def load_func(name, rtype, args):
            func = getattr(self.security, name)
            func.restype = rtype
            func.argtypes = args

        load_func("SecItemCopyMatching", c_int32, [c_void_p, POINTER(c_void_p)])
        load_func("SecItemDelete", c_int32, [c_void_p])
        load_func("SecKeyCreateSignature", c_void_p, [c_void_p, c_void_p, c_void_p, POINTER(c_void_p)])
        load_func("SecKeyIsAlgorithmSupported", c_ubyte, [c_void_p, CFIndex, c_void_p])
        load_func("SecAccessControlCreateWithFlags", c_void_p, [CFAllocatorRef, c_void_p, c_ulong, POINTER(c_void_p)])
        load_func("SecKeyCreateRandomKey", c_void_p, [c_void_p, POINTER(c_void_p)])
        load_func("SecKeyCopyPublicKey", c_void_p, [c_void_p])
        load_func("SecKeyCopyExternalRepresentation", c_void_p, [c_void_p, POINTER(c_void_p)])
        load_func("SecItemAdd", c_int32, [c_void_p, POINTER(c_void_p)])
        load_func("SecItemUpdate", c_int32, [c_void_p, c_void_p])
        load_func("SecAccessCreate", c_int32, [c_void_p, c_void_p, POINTER(c_void_p)])
        load_func("SecACLCreateWithSimpleContents",
                  c_int32,
                  [c_void_p, c_void_p, c_void_p, c_uint16, POINTER(c_void_p)])

        attr_list = ["kSecClass", "kSecClassKey", "kSecAttrLabel", "kSecAttrKeyClass", "kSecAttrKeyClassPrivate",
                     "kSecReturnRef", "kSecReturnAttributes", "kSecValueRef", "kSecAttrKeyType",
                     "kSecAttrKeySizeInBits", "kSecValueRef", "kSecAttrKeyType", "kSecAttrKeySizeInBits",
                     "kSecAttrKeyTypeRSA", "kSecAttrKeyTypeECSECPrimeRandom",
                     "kSecKeyAlgorithmRSASignatureDigestPKCS1v15SHA256",
                     "kSecKeyAlgorithmECDSASignatureDigestX962SHA256", "kSecAttrTokenID",
                     "kSecAttrTokenIDSecureEnclave", "kSecPrivateKeyAttrs", "kSecAttrIsPermanent",
                     "kSecAttrAccessControl", "kSecAttrAccess", "kSecAttrAccessibleAfterFirstUnlock",
                     "kSecAttrApplicationTag"]
        for attr in attr_list:
            setattr(self, attr, c_void_p.in_dll(self.security, attr))
        self.kSecAccessControlPrivateKeyUsage = 1 << 30

        cf.CFRetain.restype = c_void_p
        cf.CFRetain.argtypes = [c_void_p]

        self.kCFBooleanTrue = c_void_p.in_dll(cf, "kCFBooleanTrue")

        self.errSecSuccess = 0
        self.errSecItemNotFound = -25300

    @staticmethod
    def create_data(data: bytes) -> c_void_p:
        return cf.CFDataCreate(None, data, len(data))

    @staticmethod
    def get_data(data_ptr: c_void_p):
        length = cf.CFDataGetLength(data_ptr)
        array = (c_ubyte * length)()
        cf.CFDataGetBytes(data_ptr, CFRange(0, length), array)
        return bytes(array)


class ReleaseGuard:
    def __init__(self):
        self.__dict__['_dict'] = {}

    def __getattr__(self, item):
        return self._dict[item]

    def __setattr__(self, key, value):
        self._dict[key] = value

    def __enter__(self):
        if len(self._dict) > 0:
            raise RuntimeError('Re-enter a ReleaseGuard')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for v in self._dict.values():
            if v:
                cf.CFRelease(v)
        return False



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
import logging
from hashlib import sha256
from ...encoding import FormalName, SignatureType, Name, SignaturePtrs
from ...types import Validator


async def sha256_digest_checker(name: FormalName, sig: SignaturePtrs) -> bool:
    sig_info = sig.signature_info
    covered_part = sig.signature_covered_part
    sig_value = sig.signature_value_buf
    if sig_info and sig_info.signature_type == SignatureType.DIGEST_SHA256:
        sha256_algo = sha256()
        if not covered_part or not sig_value:
            ret = False
        else:
            for blk in covered_part:
                sha256_algo.update(blk)
            ret = sha256_algo.digest() == sig_value
        logging.debug('Digest check %s -> %s' % (Name.to_str(name), ret))
        return ret
    else:
        return True


# This is automatically called
async def params_sha256_checker(name: FormalName, sig: SignaturePtrs) -> bool:
    covered_part = sig.digest_covered_part
    sig_value = sig.digest_value_buf
    sha256_algo = sha256()
    if not covered_part or not sig_value:
        ret = False
    else:
        for blk in covered_part:
            sha256_algo.update(blk)
        ret = sha256_algo.digest() == sig_value
    logging.debug('Interest params-sha256 check %s -> %s' % (Name.to_str(name), ret))
    return ret


def union_checker(*args) -> Validator:
    async def wrapper(name: FormalName, sig: SignaturePtrs) -> bool:
        for checker in args:
            if not await checker(name, sig):
                return False
        return True
    return wrapper



#!/usr/bin/env python3

import io
import re
from setuptools import setup, find_packages  # Always prefer setuptools over distutils


with io.open("src/ndn/__init__.py", "rt", encoding="utf8") as f:
    version = re.search(r'__version__ = "(.*?)"', f.read()).group(1)


requirements = ['pycryptodomex >= 3.9.9', 'pygtrie >= 2.3.3', 'dataclasses >= 0.6']
setup(
    name='python-ndn',
    version=version,
    description='An NDN client library with AsyncIO support in Python 3',
    url='https://github.com/named-data/python-ndn',
    author='Xinyu Ma',
    author_email='ma.xinyu.26a@kyoto-u.jp',
    download_url='https://pypi.python.org/pypi/python-ndn',
    project_urls={
        "Bug Tracker": "https://github.com/named-data/python-ndn/issues",
        "Documentation": "https://python-ndn.readthedocs.io/",
        "Source Code": "https://github.com/named-data/python-ndn",
    },
    license='Apache License 2.0',
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',

        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'Topic :: Internet',
        'Topic :: System :: Networking',

        'License :: OSI Approved :: Apache Software License',

        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],

    keywords='NDN',

    packages=find_packages('src'),
    package_dir={'': 'src'},

    install_requires=requirements,
    python_requires=">=3.6",
    extras_require={
        "dev":  ["pytest>=5.3.5", "pytest-cov>=2.8.1", "flake8>=3.7.9"],
    }
)



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
# TODO: Change these names
from .schema_tree import Node
from .util import norm_pattern
from ..encoding import Name, Component, TlvModel, NameField, ContentType
from ..types import InterestTimeout
from ..utils import timestamp


class LocalResource(Node):
    """
    LocalResource is a custom node that preloads some data.
    When need() is called, it returns the loaded data directly.
    This node type does not interact with the network.
    """
    def __init__(self, parent=None, data=None):
        super().__init__(parent)
        self.data = data

    async def on_register(self, root, app, prefix, cached: bool):
        self.prefix = prefix
        self.app = app
        return await app.register(prefix, root._on_interest_root, root._int_validator, True)

    async def need(self, match, **kwargs):
        return self.data

    async def provide(self, match, content, **kwargs):
        self.data = content


class SegmentedNode(Node):
    """
    SegmentedNode represents a segmented object.
    The segmented object is composed with multiple Data packets,
    whose name have a suffix "/seg=seg_no" attached to the object's name.
    The ``provide`` function handles segmentation, and the ``need`` function handles reassembly.

    .. note::

        Currently, the fetching pipeline is a simple one-by-one pipeline.
        where only one Interest will be in-flight at one time.
    """
    SEGMENT_PATTERN = norm_pattern('<seg:seg_no>')[0]
    SEGMENT_SIZE = 4400

    def __init__(self, parent=None, timeout=4000, retry_times=3, segment_size=SEGMENT_SIZE):
        super().__init__(parent)
        self._set(self.SEGMENT_PATTERN, Node())
        self.timeout = timeout
        self.retry_times = retry_times
        self.segment_size = segment_size

    async def retry(self, submatch, must_be_fresh):
        trial_times = 0
        while True:
            try:
                return await submatch.need(must_be_fresh=must_be_fresh, lifetime=self.timeout, can_be_prefix=False)
            except InterestTimeout:
                trial_times += 1
                if trial_times >= self.retry_times:
                    raise

    async def need(self, match, **kwargs):
        if match.pos < len(match.name):
            raise ValueError(f'{Name.to_str(match.name)} does not match with the structure')
        subname = match.name + [None]
        must_be_fresh = kwargs.get('must_be_fresh', True)
        contents = []
        cur = 0
        while True:
            subname[-1] = Component.from_segment(cur)
            submatch = match.finer_match(subname)
            segment, meta_data = await self.retry(submatch, must_be_fresh)
            contents.append(segment)
            if meta_data['final_block_id'] == subname[-1]:
                break
            cur += 1
        ret = b''.join(contents)
        meta_data_ret = {
            **match.env,
            'content_type': meta_data['content_type'],
            'block_count': cur + 1,
            'freshness_period': meta_data['freshness_period']
        }
        return ret, meta_data_ret

    async def provide(self, match, content, **kwargs):
        seg_cnt = (len(content) + self.segment_size - 1) // self.segment_size
        subname = match.name + [None]
        final_block_id = Component.from_segment(seg_cnt - 1)
        for i in range(seg_cnt):
            subname[-1] = Component.from_segment(i)
            submatch = match.finer_match(subname)
            kwargs['final_block_id'] = final_block_id
            await submatch.provide(content[i*self.segment_size:(i+1)*self.segment_size], **kwargs)

    async def process_int(self, match, param, app_param, raw_packet):
        if match.pos == len(match.name):
            submatch = match.finer_match(match.name + [Component.from_segment(0)])
            return await submatch.on_interest(param, None, raw_packet)


class RDRNode(Node):
    """
    RDRNode represents a versioned and segmented object whose encoding follows the RDR protocol.
    Its ``provide`` function generates the metadata packet, and ``need`` function handles version discovery.
    """
    class MetaDataValue(TlvModel):
        name = NameField()

    class MetaData(Node):
        VERSION_PATTERN = norm_pattern('<v:timestamp>')[0]
        FRESHNESS_PERIOD = 10

        def __init__(self, parent=None):
            super().__init__(parent)
            self._set(self.VERSION_PATTERN, Node(self))

        def make_metadata(self, match):
            metadata = RDRNode.MetaDataValue()
            metadata.name = match.name[:-1] + [Component.from_version(self.parent.timestamp)]
            return metadata.encode()

        async def process_int(self, match, param, app_param, raw_packet):
            if match.pos == len(match.name) and self.parent.timestamp is not None and param.can_be_prefix:
                metaname = match.name + [Component.from_version(timestamp())]
                submatch = match.finer_match(metaname)
                await submatch.put_data(self.make_metadata(match), send_packet=True,
                                        freshness_period=self.FRESHNESS_PERIOD)

        async def need(self, match, **kwargs):
            if self.parent.timestamp is None:
                return await super().need(match, **kwargs)
            else:
                meta_info = {
                    **match.env,
                    'content_type': ContentType.BLOB,
                    'freshness_period': self.FRESHNESS_PERIOD,
                    'final_block_id': None
                }
                return self.make_metadata(match), meta_info

    def __init__(self, parent=None, **kwargs):
        super().__init__(parent)
        self['/32=metadata'] = RDRNode.MetaData(self)
        self['<v:timestamp>'] = SegmentedNode(self, **kwargs)
        self.timestamp = None

    async def need(self, match, **kwargs):
        submatch = match.finer_match(match.name + [Component.from_str('32=metadata')])
        lifetime = kwargs.get('lifetime', None)
        meta_int_param = {'lifetime': lifetime} if lifetime else {}
        metadata_val, _ = await submatch.need(must_be_fresh=True, can_be_prefix=True, **meta_int_param)
        metadata = RDRNode.MetaDataValue.parse(metadata_val, ignore_critical=True)

        submatch = match.finer_match(metadata.name)
        return await submatch.need(**kwargs)

    async def provide(self, match, content, **kwargs):
        self.timestamp = timestamp()
        submatch = match.finer_match(match.name + [Component.from_version(self.timestamp)])
        await submatch.provide(content, **kwargs)



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
import sys
import logging
from typing import Tuple
from ctypes import c_void_p, pointer, c_int
from Cryptodome.Hash import SHA256
from Cryptodome.PublicKey import RSA, ECC
from ...encoding import Signer, NonStrictName, Name, SignatureType, KeyLocator, BinaryStr, FormalName
from .tpm import Tpm
if sys.platform == 'darwin':
    from ...contrib.cocoapy import cf, CFSTR, ObjCInstance, cfstring_to_string, cfnumber_to_number, kCFNumberIntType
    from ..osx import OsxSec, ReleaseGuard


class OsxSigner(Signer):
    def __init__(self, key_name: NonStrictName, key_bits, key_type, key_ref):
        self.key_ref = key_ref
        self.key_name = key_name
        if key_type == cfstring_to_string(OsxSec().kSecAttrKeyTypeRSA):
            self.key_type = SignatureType.SHA256_WITH_RSA
            self.key_size = key_bits // 8
        elif key_type == cfstring_to_string(OsxSec().kSecAttrKeyTypeECSECPrimeRandom):
            self.key_type = SignatureType.SHA256_WITH_ECDSA
            self.key_size = (key_bits * 2 + 7) // 8
            self.key_size += self.key_size % 2
            self.key_size += 8
        else:
            raise ValueError(f'Unrecognized key type {key_type}')

    def write_signature_info(self, signature_info):
        signature_info.signature_type = self.key_type
        signature_info.key_locator = KeyLocator()
        signature_info.key_locator.name = self.key_name

    def get_signature_value_size(self):
        return self.key_size

    def __del__(self):
        cf.CFRelease(self.key_ref)

    def write_signature_value(self, wire, contents) -> int:
        h = SHA256.new()
        for blk in contents:
            h.update(blk)

        # KeyDigest
        sec = OsxSec()
        with ReleaseGuard() as g:
            g.digest = sec.create_data(h.digest())
            error = c_void_p()
            if self.key_type == SignatureType.SHA256_WITH_RSA:
                algo = sec.kSecKeyAlgorithmRSASignatureDigestPKCS1v15SHA256
            else:
                algo = sec.kSecKeyAlgorithmECDSASignatureDigestX962SHA256
            g.sig = sec.security.SecKeyCreateSignature(self.key_ref, algo, g.digest, pointer(error))
            if error.value is not None:
                raise RuntimeError('Create signature failed')
            signature = sec.get_data(g.sig)
        real_len = len(signature)
        wire[:real_len] = signature
        return real_len


class TpmOsxKeychain(Tpm):
    @staticmethod
    def _get_key(key_name: NonStrictName):
        sec = OsxSec()
        with ReleaseGuard() as g:
            logging.debug('Get OSX Key %s' % Name.to_str(key_name))
            g.key_label = CFSTR(Name.to_str(key_name))
            g.query = ObjCInstance(cf.CFDictionaryCreateMutable(None, 6, cf.kCFTypeDictionaryKeyCallBacks, None))
            cf.CFDictionaryAddValue(g.query, sec.kSecClass, sec.kSecClassKey)
            cf.CFDictionaryAddValue(g.query, sec.kSecAttrLabel, g.key_label)
            cf.CFDictionaryAddValue(g.query, sec.kSecAttrKeyClass, sec.kSecAttrKeyClassPrivate)
            cf.CFDictionaryAddValue(g.query, sec.kSecReturnAttributes, sec.kCFBooleanTrue)
            cf.CFDictionaryAddValue(g.query, sec.kSecReturnRef, sec.kCFBooleanTrue)

            g.dic = c_void_p()
            ret = sec.security.SecItemCopyMatching(g.query, pointer(g.dic))
            if ret == sec.errSecItemNotFound:
                raise KeyError(f"Unable to find key {key_name}")
            elif ret != sec.errSecSuccess:
                raise RuntimeError(f"Error happened when searching specific key {key_name}")

            key_type = cfstring_to_string(cf.CFDictionaryGetValue(g.dic, sec.kSecAttrKeyType))
            key_bits = cfnumber_to_number(cf.CFDictionaryGetValue(g.dic, sec.kSecAttrKeySizeInBits))
            key_ref = cf.CFRetain(cf.CFDictionaryGetValue(g.dic, sec.kSecValueRef))
        return key_type, key_bits, key_ref

    def get_signer(self, key_name: NonStrictName) -> Signer:
        key_type, key_bits, key_ref = self._get_key(key_name)
        return OsxSigner(key_name, key_bits, key_type, key_ref)

    def key_exist(self, key_name: FormalName) -> bool:
        try:
            self._get_key(key_name)
            return True
        except KeyError:
            return False

    def delete_key(self, key_name: FormalName):
        sec = OsxSec()
        with ReleaseGuard() as g:
            logging.debug('Delete OSX Key %s' % Name.to_str(key_name))
            g.key_label = CFSTR(Name.to_str(key_name))
            g.query = ObjCInstance(cf.CFDictionaryCreateMutable(None, 3, cf.kCFTypeDictionaryKeyCallBacks, None))
            cf.CFDictionaryAddValue(g.query, sec.kSecClass, sec.kSecClassKey)
            cf.CFDictionaryAddValue(g.query, sec.kSecAttrLabel, g.key_label)
            sec.security.SecItemDelete(g.query)

    @staticmethod
    def _convert_key_format(key_bits: BinaryStr, key_type: str):
        if key_type == 'rsa':
            return RSA.import_key(key_bits).export_key(format='DER')
        elif key_type == 'ec':
            xp = int.from_bytes(key_bits[1:33], 'big')
            yp = int.from_bytes(key_bits[33:], 'big')
            return ECC.construct(curve='P-256', point_x=xp, point_y=yp).export_key(format='DER')
        else:
            raise ValueError(f'Unsupported key type {key_type}')

    def generate_key(self, id_name: FormalName, key_type: str = 'rsa', **kwargs) -> Tuple[FormalName, BinaryStr]:
        sec = OsxSec()
        with ReleaseGuard() as g:
            logging.debug('Generating OSX Key %s' % key_type)

            # Get key type and size
            if key_type == 'rsa':
                mac_key_type = sec.kSecAttrKeyTypeRSA
                key_size = c_int(kwargs.pop('key_size', 2048))
            elif key_type == 'ec':
                mac_key_type = sec.kSecAttrKeyTypeECSECPrimeRandom
                key_size = kwargs.pop('key_size', 256)
                if key_size != 256:
                    raise ValueError(f'Unsupported ECC curve P-{key_size}. Please use P-256.')
                key_size = c_int(key_size)
            else:
                raise ValueError(f'Unsupported key type {key_type}')
            g.key_size = cf.CFNumberCreate(None, kCFNumberIntType, pointer(key_size))

            # Generate private key
            g.pk_attr = ObjCInstance(cf.CFDictionaryCreateMutable(None, 1, cf.kCFTypeDictionaryKeyCallBacks, None))
            cf.CFDictionaryAddValue(g.pk_attr, sec.kSecAttrIsPermanent, sec.kCFBooleanTrue)

            g.attr = ObjCInstance(cf.CFDictionaryCreateMutable(None, 3, cf.kCFTypeDictionaryKeyCallBacks, None))
            cf.CFDictionaryAddValue(g.attr, sec.kSecAttrKeyType, mac_key_type)
            cf.CFDictionaryAddValue(g.attr, sec.kSecAttrKeySizeInBits, g.key_size)
            cf.CFDictionaryAddValue(g.attr, sec.kSecPrivateKeyAttrs, g.pk_attr)

            error = c_void_p()
            g.pri_key = sec.security.SecKeyCreateRandomKey(g.attr, pointer(error))
            if not g.pri_key:
                raise RuntimeError('Unable to create specific key')

            # Derive public key
            g.pub_key = sec.security.SecKeyCopyPublicKey(g.pri_key)
            g.pub_key_der = sec.security.SecKeyCopyExternalRepresentation(g.pub_key, pointer(error))
            if not g.pub_key_der:
                raise RuntimeError('Unable to export the public key')
            pub_key = self._convert_key_format(sec.get_data(g.pub_key_der), key_type)

            # Construct key name
            key_name = self.construct_key_name(id_name, pub_key, **kwargs)
            key_name_str = Name.to_str(key_name)
            g.key_label = CFSTR(Name.to_str(key_name_str))
            logging.debug('Generated OSX Key %s' % key_name_str)

            # SecItemUpdate: kSecAttrLabel, kSecAttrAccessControl
            g.query = ObjCInstance(cf.CFDictionaryCreateMutable(None, 2, cf.kCFTypeDictionaryKeyCallBacks, None))
            cf.CFDictionaryAddValue(g.query, sec.kSecValueRef, g.pri_key)
            cf.CFDictionaryAddValue(g.query, sec.kSecClass, sec.kSecClassKey)

            g.update = ObjCInstance(cf.CFDictionaryCreateMutable(None, 2, cf.kCFTypeDictionaryKeyCallBacks, None))
            cf.CFDictionaryAddValue(g.update, sec.kSecAttrLabel, g.key_label)
            # Access cannot be changed in this way
            # g.access = c_void_p()
            # sec.security.SecAccessCreate(g.key_label, None, pointer(g.access))
            # g.acl = c_void_p()
            # sec.security.SecACLCreateWithSimpleContents(g.access, None, g.key_label, 0, pointer(g.acl))
            # cf.CFDictionaryAddValue(g.update, sec.kSecAttrAccess, g.access)

            ret = sec.security.SecItemUpdate(g.query, g.update)
            if ret != sec.errSecSuccess:
                raise RuntimeError('Unable to set the key label and access')
        return key_name, pub_key



from .keychain import Keychain
from .keychain_sqlite3 import Identity, Key, Certificate, KeychainSqlite3
from .keychain_digest import KeychainDigest

__all__ = ['Keychain', 'Identity', 'Key', 'Certificate', 'KeychainSqlite3', 'KeychainDigest']



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
import sqlite3
from typing import Iterator
from dataclasses import dataclass
from typing import Dict, Any, Mapping
from ...encoding import FormalName, BinaryStr, NonStrictName, Name
from ...app_support.security_v2 import self_sign
from ..signer.sha256_digest_signer import DigestSha256Signer
from ..tpm.tpm import Tpm
from .keychain import Keychain


@dataclass
class Certificate:
    """
    A dataclass for a Certificate.

    :ivar id: its id in the database.
    :vartype id: int
    :ivar key: the Name of the associated Key.
    :vartype key: :any:`FormalName`
    :ivar name: its Name.
    :vartype name: :any:`FormalName`
    :ivar data: the content.
    :vartype data: bytes
    :ivar is_default: whether this is the default Identity.
    :vartype is_default: bool
    """
    id: int
    key: FormalName
    name: FormalName
    data: BinaryStr
    is_default: bool


class Key(Mapping[FormalName, Certificate]):
    """
    A Key. It behaves like an immutable ``dict`` from :any:`FormalName` to :any:`Certificate`.

    :ivar row_id: its id in the database.
    :vartype row_id: int
    :ivar identity: the Name of the associated Identity.
    :vartype identity: :any:`FormalName`.
    :ivar name: its Name.
    :vartype name: :any:`FormalName`
    :ivar key_bits: the key bits of the public key.
    :vartype key_bits: bytes
    :ivar is_default: whether this is the default Identity.
    :vartype is_default: bool
    """
    row_id: int
    identity: FormalName
    name: FormalName
    key_bits: BinaryStr
    is_default: bool

    def __init__(self, pib, identity, row_id, name, key_bits, is_default):
        self.pib = pib
        self.identity = identity
        self.row_id = row_id
        self.name = name
        self.key_bits = key_bits
        self.is_default = is_default

    def __len__(self) -> int:
        cursor = self.pib.conn.execute('SELECT count(*) FROM keys WHERE identity_id=?', (self.row_id,))
        ret = cursor.fetchone()[0]
        cursor.close()
        return ret

    def __getitem__(self, name: NonStrictName) -> Certificate:
        name = Name.to_bytes(name)
        sql = 'SELECT id, certificate_name, certificate_data, is_default FROM certificates WHERE certificate_name=?'
        cursor = self.pib.conn.execute(sql, (name,))
        data = cursor.fetchone()
        if not data:
            raise KeyError(name)
        row_id, cert_name, cert_data, is_default = data
        cursor.close()
        return Certificate(id=row_id, key=self.name, name=cert_name, data=cert_data, is_default=is_default != 0)

    def __iter__(self) -> Iterator[FormalName]:
        cursor = self.pib.conn.execute('SELECT certificate_name FROM certificates WHERE key_id=?', (self.row_id,))
        while True:
            name = cursor.fetchone()
            if not name:
                break
            yield Name.from_bytes(name[0])
        cursor.close()

    def del_cert(self, name: NonStrictName):
        """
        Delete a specific Certificare.

        :param name: the Name of the Key to delete.
        :type name: :any:`NonStrictName`
        """
        return self.pib.del_certificate(name)

    def has_default_cert(self) -> bool:
        """
        Whether it has a default Certificate.

        :return: ``True`` if there is one.
        """
        cursor = self.pib.conn.execute('SELECT id FROM certificates WHERE is_default=1 AND key_id=?', (self.row_id,))
        ret = cursor.fetchone() is not None
        cursor.close()
        return ret

    def set_default_cert(self, name: NonStrictName):
        """
        Set the default Certificate.

        :param name: the Name of the new default Certificate.
        :type name: :any:`NonStrictName`
        """
        name = Name.to_bytes(name)
        self.pib.conn.execute('UPDATE certificates SET is_default=1 WHERE certificate_name=?', (name,))
        self.pib.conn.commit()

    def default_cert(self) -> Certificate:
        """
        Get the default Certificate.

        :return: the default Certificate.
        """
        sql = ('SELECT id, certificate_name, certificate_data, is_default '
               'FROM certificates WHERE is_default=1 AND key_id=?')
        cursor = self.pib.conn.execute(sql, (self.row_id,))
        data = cursor.fetchone()
        if not data:
            raise KeyError('No default certificate')
        row_id, cert_name, cert_data, is_default = data
        cursor.close()
        return Certificate(id=row_id, key=self.name, name=cert_name, data=cert_data, is_default=is_default != 0)


class Identity(Mapping[FormalName, Key]):
    """
    An Identity. It behaves like an immutable ``dict`` from :any:`FormalName` to :any:`Key`.

    :ivar row_id: its id in the database.
    :vartype row_id: int
    :ivar name: its Name.
    :vartype name: :any:`FormalName`
    :ivar is_default: whether this is the default Identity.
    :vartype is_default: bool
    """
    row_id: int
    name: FormalName
    is_default: bool

    def __init__(self, pib, row_id, name, is_default):
        self.pib = pib
        self.row_id = row_id
        self.name = name
        self.is_default = is_default

    def __len__(self) -> int:
        cursor = self.pib.conn.execute('SELECT count(*) FROM keys WHERE identity_id=?', (self.row_id,))
        ret = cursor.fetchone()[0]
        cursor.close()
        return ret

    def __getitem__(self, name: NonStrictName) -> Key:
        name = Name.to_bytes(name)
        cursor = self.pib.conn.execute('SELECT id, key_name, key_bits, is_default FROM keys WHERE key_name=?',
                                       (name,))
        data = cursor.fetchone()
        if not data:
            raise KeyError(name)
        row_id, key_name, key_bits, is_default = data
        cursor.close()
        return Key(self.pib, self.name, row_id, Name.from_bytes(key_name), key_bits, is_default != 0)

    def __iter__(self) -> Iterator[FormalName]:
        cursor = self.pib.conn.execute('SELECT key_name FROM keys WHERE identity_id=?', (self.row_id,))
        while True:
            name = cursor.fetchone()
            if not name:
                break
            yield Name.from_bytes(name[0])
        cursor.close()

    def del_key(self, name: NonStrictName):
        """
        Delete a specific Key.

        :param name: the Name of the Key to delete.
        :type name: :any:`NonStrictName`
        """
        return self.pib.del_key(name)

    def new_key(self, key_type: str) -> Key:
        """
        Create a new key with default arguments.

        :param key_type: the type of the Key. Can be ``ec`` or ``rsa``.
        :return: the new Key.
        """
        return self.pib.new_key(self.name, key_type=key_type)

    def has_default_key(self) -> bool:
        """
        Whether it has a default Key.

        :return: ``True`` if there is one.
        """
        cursor = self.pib.conn.execute('SELECT id FROM keys WHERE is_default=1 AND identity_id=?', (self.row_id,))
        ret = cursor.fetchone() is not None
        cursor.close()
        return ret

    def set_default_key(self, name: NonStrictName):
        """
        Set the default Key.

        :param name: the Name of the new default Key.
        :type name: :any:`NonStrictName`
        """
        name = Name.to_bytes(name)
        self.pib.conn.execute('UPDATE keys SET is_default=1 WHERE key_name=?', (name,))
        self.pib.conn.commit()

    def default_key(self) -> Key:
        """
        Get the default Key.

        :return: the default Key.
        """
        sql = 'SELECT id, key_name, key_bits, is_default FROM keys WHERE is_default=1 AND identity_id=?'
        cursor = self.pib.conn.execute(sql, (self.row_id,))
        data = cursor.fetchone()
        if not data:
            raise KeyError('No default key')
        row_id, key_name, key_bits, is_default = data
        cursor.close()
        return Key(self.pib, self.name, row_id, Name.from_bytes(key_name), key_bits, is_default != 0)


class KeychainSqlite3(Keychain):
    r"""
    Store public infomation in a Sqlite3 database and private keys in a TPM.

    :ivar path: the path to the database. The default path is ``~/.ndn/pib.db``.
    :vartype path: str
    :ivar tpm: an instance of TPM.
    :vartype tpm: :class:`Tpm`
    :ivar tpm_locator: a URI string describing the location of TPM.
    :vartype tpm_locator: str
    """
    tpm: Tpm
    path: str
    tpm_locator: str
    _signer_cache: dict

    def __init__(self, path: str, tpm: Tpm):
        self.path = path
        self.conn = sqlite3.connect(path)
        cursor = self.conn.execute('SELECT tpm_locator FROM tpmInfo')
        self.tpm_locator = cursor.fetchone()[0]
        cursor.close()
        self.tpm = tpm
        self._signer_cache = {}

    def __iter__(self) -> Iterator[FormalName]:
        cursor = self.conn.execute('SELECT identity FROM identities')
        while True:
            name = cursor.fetchone()
            if not name:
                break
            yield Name.from_bytes(name[0])
        cursor.close()

    def __len__(self) -> int:
        cursor = self.conn.execute('SELECT count(*) FROM identities')
        ret = cursor.fetchone()[0]
        cursor.close()
        return ret

    def __getitem__(self, name: NonStrictName) -> Identity:
        name = Name.to_bytes(name)
        cursor = self.conn.execute('SELECT id, identity, is_default FROM identities WHERE identity=?', (name,))
        data = cursor.fetchone()
        if not data:
            raise KeyError(name)
        row_id, identity, is_default = data
        cursor.close()
        return Identity(self, row_id, Name.from_bytes(identity), is_default != 0)

    def has_default_identity(self) -> bool:
        """
        Whether there is a default Identity.
        :return: ``True`` if there is one.
        """
        cursor = self.conn.execute('SELECT id FROM identities WHERE is_default=1')
        ret = cursor.fetchone() is not None
        cursor.close()
        return ret

    def set_default_identity(self, name: NonStrictName):
        """
        Set the default Identity.

        :param name: the Name of the new default Identity.
        :type name: :any:`NonStrictName`
        """
        name = Name.to_bytes(name)
        self.conn.execute('UPDATE identities SET is_default=1 WHERE identity=?', (name,))
        self.conn.commit()

    def default_identity(self) -> Identity:
        """
        Get the default Identity.

        :return: the default Identity.
        """
        cursor = self.conn.execute('SELECT id, identity, is_default FROM identities WHERE is_default=1')
        data = cursor.fetchone()
        if not data:
            raise KeyError('No default identity')
        row_id, identity, is_default = data
        cursor.close()
        return Identity(self, row_id, Name.from_bytes(identity), is_default != 0)

    def new_identity(self, name: NonStrictName) -> Identity:
        """
        Create a new Identity without a default Key.
        This is used to control the Keychain in a fine-grained way.

        :param name: the Name of the new Identity.
        :type name: :any:`NonStrictName`
        :return: the Identity created.
        """
        name = Name.to_bytes(name)
        if name not in self:
            self.conn.execute('INSERT INTO identities (identity) VALUES (?)', (name,))
            self.conn.commit()
        else:
            raise KeyError(f'Identity {Name.to_str(name)} already exists')
        if not self.has_default_identity():
            self.set_default_identity(name)
        return self[name]

    def touch_identity(self, id_name: NonStrictName) -> Identity:
        """
        Get an Identity with specific name. Create a new one if it does not exist.
        The newly created one will automatically have a default ECC Key and self-signed Certificate.

        :param id_name: the Name of Identity.
        :type id_name: :any:`NonStrictName`
        :return: the specified Identity.
        """
        name = Name.to_bytes(id_name)
        if name not in self:
            self.conn.execute('INSERT INTO identities (identity) VALUES (?)', (name,))
            self.conn.commit()
            self.new_key(name)
        if not self.has_default_identity():
            self.set_default_identity(name)
        return self[name]

    def shutdown(self):
        """
        Close the connection.
        """
        self.conn.close()

    def del_identity(self, name: NonStrictName):
        """
        Delete a specific Identity.

        :param name: the Identity Name.
        :type name: :any:`NonStrictName`
        """
        name = Name.to_bytes(name)
        for key_name in self[name]:
            self.tpm.delete_key(key_name)
        self.conn.execute('DELETE FROM identities WHERE identity=?', (name,))
        self.conn.commit()
        self._signer_cache = {}

    def get_signer(self, sign_args: Dict[str, Any]):
        if sign_args.pop('no_signature', False):
            return None
        if sign_args.pop('digest_sha256', False):
            return DigestSha256Signer()
        key_name = sign_args.pop('key', None)
        if not key_name:
            id_name = sign_args.pop('identity', None)
            if id_name:
                if isinstance(id_name, Identity):
                    identity = id_name
                else:
                    identity = self[id_name]
            else:
                identity = self.default_identity()
            key_name = identity.default_key().name
        elif isinstance(key_name, Key):
            key_name = key_name.name
        key_name_bytes = Name.to_bytes(key_name)
        signer = self._signer_cache.get(key_name_bytes, None)
        if not signer:
            signer = self.tpm.get_signer(key_name)
            self._signer_cache[key_name_bytes] = signer
        return signer

    def del_key(self, name: NonStrictName):
        """
        Delete a specific Key.

        :param name: the Key Name.
        :type name: :any:`NonStrictName`
        """
        formal_name = Name.normalize(name)
        name = Name.to_bytes(name)
        self.conn.execute('DELETE FROM keys WHERE key_name=?', (name,))
        self.conn.commit()
        self.tpm.delete_key(formal_name)
        self._signer_cache = {}

    def del_cert(self, name: NonStrictName):
        """
        Delete a specific Certificate.

        :param name: the Certificate Name.
        :type name: :any:`NonStrictName`
        """
        name = Name.to_bytes(name)
        self.conn.execute('DELETE FROM certificates WHERE certificate_name=?', (name,))
        self.conn.commit()
        self._signer_cache = {}

    def new_key(self, id_name: NonStrictName, key_type: str = 'ec', **kwargs) -> Key:
        """
        Generate a new key for a specific Identity.

        :param id_name: the Name of Identity.
        :type id_name: :any:`NonStrictName`
        :param key_type: the type of key. Can be one of the following:

            + ``ec``: ECDSA key.
            + ``rsa``: RSA key.

        :param kwargs: keyword arguments.

        :Keyword Arguments:

            + **key_size** (:class:`int`) - key size in bit.
            + **key_id** (Union[:any:`BinaryStr`, :class:`str`]) - a one-Component ID of the Key.
            + **key_id_type** (:class:`str`) - the method to generate the ID if *key_id* is not specified.
              Can be ``random`` or ``sha256``.

        :return: the new Key.
        """
        name = Name.normalize(id_name)
        if name not in self:
            raise KeyError(f'Identity {Name.to_str(id_name)} does not exist')
        identity = self[name]
        key_name, pub_key = self.tpm.generate_key(name, key_type, **kwargs)
        signer = self.tpm.get_signer(key_name)
        cert_name, cert_data = self_sign(key_name, pub_key, signer)
        key_name = Name.to_bytes(key_name)
        cert_name = Name.to_bytes(cert_name)
        self.conn.execute('INSERT INTO keys (identity_id, key_name, key_bits) VALUES (?, ?, ?)',
                          (identity.row_id, key_name, pub_key))
        self.conn.execute('INSERT INTO certificates (key_id, certificate_name, certificate_data)'
                          'VALUES ((SELECT id FROM keys WHERE key_name=?), ?, ?)',
                          (key_name, cert_name, bytes(cert_data)))
        self.conn.commit()

        if not identity.has_default_key():
            identity.set_default_key(key_name)
        return identity[key_name]



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
import abc
import collections
from typing import Dict, Any


class Keychain(collections.abc.Mapping):
    """
    The abstract Keychain class, derived from :class:`collections.abc.Mapping`.
    It behaves like an immutable dict from :any:`FormalName` to Identity.
    The implementation of Identity varies with concrete implementations.
    Generally, its methods should also accept :any:`NonStrictName` as inputs.
    This includes operators such as ``in`` and ``[]``.
    """
    # __getitem__ will be called extra times, but there is no need to optimize for performance
    @abc.abstractmethod
    def get_signer(self, sign_args: Dict[str, Any]):
        """
        Get a signer from sign_args.

        :param sign_args: the signing arguments provided by the application.
        :return: a signer.
        :rtype: :any:`Signer`
        """
        pass



from .keychain import *
from .signer import *
from .tpm import *
from .validator import *


__all__ = []
__all__.extend(keychain.__all__)
__all__.extend(signer.__all__)
__all__.extend(tpm.__all__)
__all__.extend(validator.__all__)
